class LevelType:
    FLAT = "flat"
    DEFAULT = "default"

class GameMode:
    SURVIVAL = "survival"
    CREATIVE = "creative"
    SPECTATOR = "spectator"
    ADVENTURE = "adventure"

class Difficulty:
    PEACEFUL = "peaceful"
    EASY = "easy"
    NORMAL = "normal"
    HARD = "hard"

class Dimension:
    NETHER = "the_nether"
    OVERWORLD = "overworld"
    END = "the_end"

class BrandChannel:
    PREFIXED = "MC|Brand"
    IDENTIFIER = "minecraft:brand"

class ChatSetting:
    ENABLED = "enabled"
    COMMANDS_ONLY = "commandsOnly"
    DISABLED = "disabled"

class ViewDistance:
    FAR = "far"
    NORMAL = "normal"
    SHORT = "short"
    TINY = "tiny"


from javascript import require, On, Once, off, once
from javascript.proxy import Proxy
from rich.console import Console
from tinydb import TinyDB, Query
import shutil
import requests
import os
import sys
import time
import fnmatch
import re
import subprocess
import git
from typing import Callable
from importlib.metadata import version as version_checker
import dataclasses
import threading
try:
    from logger import logger
    from utils import cprop
except ImportError:
    from .logger import logger
    from .utils import cprop

User = Query()


filestruc = "/"
if os.name == 'nt':
    filestruc = "\\"

__all__ = ['Bot', 'createBot']

class GameState:
    """
    Stores information about the current game state. Should not initialize manually
    """
    def __init__(self, proxy: Proxy):
        self.proxy = proxy

    @cprop()
    def level_type(self) -> str:
        "The world generation type. Posssible values are in `types.LevelType`"

    @cprop()
    def dimension(self) -> str:
        "The current dimension. Posssible values are in `types.Dimension`"

    @cprop()
    def difficulty(self) -> str:
        "The current server difficulty. Posssible values are in `types.Difficulty`"

    @cprop()
    def game_mode(self) -> str:
        "The current game mode. Posssible values are in `types.GameMode`"

    @cprop()
    def hardcore(self) -> bool:
        "Whether the client is in hardcore mode or not"

    @cprop()
    def max_players(self) -> int:
        "The maximum number of players can be allowed on a server"

    @cprop()
    def server_brand(self) -> str:
        "the current server brand. Posssible values are in `types.BrandChannel`"

    @cprop()
    def min_y(self) -> int:
        "The minimum Y level in the world"

    @cprop()
    def height(self) -> int:
        "The height of the world"

    @cprop(proxy_name="height")
    def max_y(self) -> int:
        "The height of the world"

class TimeState:
    """
    Stores information about time. Should not initialize manually
    """
    def __init__(self, proxy: Proxy):
        self.proxy = proxy

    @cprop()
    def do_daylight_cycle(self) -> bool:
        "Whether the server does daylight cycle or not"

    @cprop()
    def big_time(self) -> int:
        "Total ticks elapsed since day 0"

    @cprop()
    def time(self) -> int:
        "Total ticks elapsed since day 0. Inaccruate. Use `TimeState.big_time`"

    @cprop()
    def time_of_day(self) -> int:
        "The current time of day, in ticks. This is used in the `/time set` command"

    @cprop()
    def day(self) -> int:
        "The number of days in a world"

    @cprop()
    def is_day(self) -> bool:
        "Whether `TimeState.time_of_day` is within 13,000 and 23,000 (AKA whether it's daytime or not)"

    @cprop()
    def moon_phase(self) -> int:
        "Current moon phase. Ranges from 0 -> 7"

    @cprop()
    def big_age(self) -> int:
        "Total ticks elapsed since day 0"

    @cprop()
    def age(self) -> int:
        "Total ticks elapsed since day 0. Inaccurate. Use `TimeState.big_age`"

class ExperienceState:
    """
    Stores information about XP. Should not initalize manually
    """
    def __init__(self, proxy: Proxy):
        self.proxy = proxy

    @cprop()
    def level(self) -> int:
        "Full levels of experience"

    @cprop()
    def points(self) -> int:
        "Total experience points"

    @cprop()
    def progress(self) -> float:
        "The progress to the next full level. Ranges from 0 -> 1 (0% -> 100%)"

class SkinPartsState:
    """
    Represents the skin parts that are visible / not visible. Should not initalize manually
    """
    def __init__(self, proxy: Proxy):
        self.proxy = proxy

    @cprop()
    def show_cape(self) -> bool:
        "Whether the cape is shown"

    @cprop()
    def show_jacket(self) -> bool:
        "Whether the jacket is shown"

    @cprop()
    def show_left_sleeve(self) -> bool:
        "Whether the left sleeve is shown"

    @cprop()
    def show_right_sleeve(self) -> bool:
        "Whether the right sleeve is shown"

    @cprop()
    def show_left_pants(self) -> bool:
        "Whether the left pant is shown"

    @cprop()
    def show_right_pants(self) -> bool:
        "Whether the right pant is shown"

    @cprop()
    def show_hat(self) -> bool:
        "Whether the hat is shown"

class SettingsState:
    """
    Represents the client settings that the server needs to know. Should not initialize manually
    """
    def __init__(self, proxy: Proxy):
        self.proxy = proxy

    @cprop()
    def chat(self) -> str:
        "The current chat settings. Possible values are in `types.ChatSetting`"

    @cprop()
    def colors_enabled(self) -> bool:
        "Whether colors are recieved in chat"

    @cprop()
    def view_distance(self) -> str | int:
        "The view distance of the client. Could be one of `types.ViewDistance` or an int"

    @cprop()
    def difficulty(self) -> str:
        "The difficulty of the client. Possible values are in `types.Difficulty`"

    @property
    def skin_parts(self) -> SkinPartsState:
        "The skin parts of the client."
        return SkinPartsState(self.proxy.skin_parts)

    @cprop()
    def enable_text_filtering(self) -> bool:
        "Unused value. Default = False"

    @cprop()
    def enable_server_listing(self) -> bool:
        "Whether the player should list in the Tablist or not"

class CreativeMode:
    """
    Provides some creative mode functionalities. Should not initalize manully
    """
    def __init__(self, proxy: Proxy):
        self.proxy = proxy

    @cprop()
    def set_inventory_slot(self) -> Callable[[int, Proxy], None]:
        "Sets the inventory slot (returns Function(slot: number, item: prismarine-item.Item))"
    
    @cprop()
    def fly_to(self) -> Callable[[Proxy], None]:
        "Fly to a location (returns Function(destination: vec3.Vec3))"
    
    @cprop()
    def start_flying(self) -> Callable[[], None]:
        "Start flying (returns Function())"

    @cprop()
    def stop_flying(self) -> Callable[[], None]:
        "Stop flying (returns Function()"

    def clear_slot(self, slot: int):
        "Clears the slot"
        self.set_inventory_slot(slot, None)

    @cprop()
    def clear_inventory(self) -> Callable[[], None]:
        "Clears the inventory (returns Function())"

@dataclasses.dataclass
class CommandContext:
    """
    Command Context for Bot Commands. Should not initialize manually
    """
    sender: str
    command: str
    full_command: str
    arguments: list
    time: float
    full_message: Proxy
    bot: 'Bot'

    def respond(self, *message, whisper = False, whisper_to: str):
        """
        Respond to the command. Use whisper_to only if whisper is True
        """
        if whisper:
            self.bot.whisper(whisper_to, *message)
        else:
            self.bot.chat(*message)



def threaded(fn):
    def wrapper(*args, **kwargs):
        thread = threading.Thread(target=fn, args=args, kwargs=kwargs)
        thread.start()
        return thread
    return wrapper

class Bot(threading.Thread):
    def __init__(
            self,
            host: str,
            *,
            auth: str = "microsoft",
            port: int = 25565,
            version: str = "false",
            password: str = "",
            respawn: bool = True,
            disableChatSigning: bool = False,
            profilesFolder: str = "",
            username: str = "lodestone",
            hideErrors: bool = True,
            logErrors: bool = True,
            keepAlive: bool = True,
            loadInternalPlugins: bool = True,
            physicsEnabled: bool = True,
            defaultChatPatterns: bool = True,
            checkTimeoutInterval: int = 60 * 10000,
            ls_disable_logs: bool = False,
            ls_enable_chat_logging: bool = False,
            ls_skip_checks: bool = False,
            ls_disable_viewer: bool = False,
            ls_stop_bot_on_death: bool = False,
            ls_debug_mode: bool = False,
            ls_viewer_port: int = 5001,
            ls_use_return: bool = False,
            ls_discord_webhook: str = None,
            ls_use_discord_forums: bool = False,
            ls_api_mode: bool = False,
            ls_plugin_list: [] = None,
            easyMcToken: str = None
            
    ):
        """
        Create the bot. Parameters in camelCase are passed into mineflayer. Parameters starting with ls_ is Lodestone specific
        """
        threading.Thread.__init__(self, daemon=True)
        if ls_debug_mode:
            os.environ["DEBUG"] = "minecraft-protocol"
        else:
            os.environ["DEBUG"] = ""

        self.local_host = host
        self.local_auth = auth
        self.local_port = port
        self.local_version = version
        self.local_password = password
        self.local_disable_chat_signing = disableChatSigning
        self.local_profiles_folder = profilesFolder
        self.local_username = username
        self.local_hide_errors = hideErrors
        self.local_log_errors = logErrors
        self.local_keep_alive = keepAlive
        self.local_load_internal_plugins = loadInternalPlugins
        self.local_respawn = respawn
        self.local_physics_enabled = physicsEnabled
        self.local_default_chat_patterns = defaultChatPatterns

        self.viewer_port = ls_viewer_port
        self.disable_logs = ls_disable_logs
        self.enable_chat_logging = ls_enable_chat_logging
        self.skip_checks = ls_skip_checks
        self.disable_viewer = ls_disable_viewer
        self.discord_webhook = ls_discord_webhook
        self.stop_bot_on_death = ls_stop_bot_on_death
        self.use_discord_forums = ls_use_discord_forums
        self.api_mode = ls_api_mode
        self.plugin_list = ls_plugin_list if ls_plugin_list else []
        self.check_timeout_interval = checkTimeoutInterval
        self.easy_mc_token = easyMcToken

        self.custom_command_prefix = "!"
        self.custom_commands = {}
        """
        {
            "cmd_name": {
                "sender": sender or None,
                "callback": Callable[[CommandContext], None]
            }
        }
        """

        self.console = Console()
        self.extra_data = {}
        self.loaded_plugins = {}
        self.loaded_events = {}
        
        
        
        _ = require("mineflayer")
        _ = require("prismarine-viewer")
        _ = require("mineflayer-collectblock")
        _ = require("minecraft-data")
        _ = require("mineflayer-pathfinder")

        if not self.skip_checks:
            self.node_version, self.pip_version, self.python_version = self.__versions_check()
        else:
            self.node_version, self.pip_version, self.python_version = "unknown", "unknown", "unknown"

        if self.discord_webhook is not None:
            pass # needs plugin

        self.mineflayer = require('mineflayer')
        self.once_with_cleanup = require('mineflayer').promise_utils
        if not self.disable_viewer:
            self.mineflayer_viewer = require('prismarine-viewer')
        self.python_command = self.__check_python_command()
        if not ls_skip_checks:
            with self.console.status("[bold]Checking for updates...\n") as status:
                status.update("[bold]Updating javascript libraries...\n")
                subprocess.run(f'{self.python_command} -m javascript --update', stdout=subprocess.DEVNULL, shell=True, input=b"\n")
                status.update("[bold]Updating pip package...\n")
                subprocess.run(f'{self.python_command} -m pip install -U lodestone', stdout=subprocess.DEVNULL, shell=True, input=b"\n")
        self.logged_in = False
        self.use_return = ls_use_return
        self.msa_status = False
        self.server_name = f"{self.local_host}".lower().replace(".", "")
        if self.enable_chat_logging:
            self.chat_database = TinyDB(f"{self.server_name}Database.json")
        self.script_directory = os.path.dirname(os.path.abspath(sys.argv[0]))
        self.bot: Proxy = self.__create_bot()
        self.proxy = self.bot
        self.msa_data = False
        self.__start()

        # loads plugins
        for plugin in self.plugin_list:
            self.load_plugin(plugin)


    # RIP def __getattr__(self, name)
    # You will be missed

    def load_plugin(self, plugin: type):
        """
        Loads a singular plugin (An uninitalized class object)

        ```python
        class Plugin:
            def __init__(self, bot: lodestone.Bot):
                print("Plugin Injected")
        ```
        """
        plugin_name = plugin.__name__
        initialized_plugin = plugin(self)
        self.loaded_plugins[plugin_name] = initialized_plugin
        self.emit("plugin_load", plugin_name)
    
    def __check_python_command(self):
        try:
            subprocess.check_output(['python', '--version'])
            return 'python'
        except:
            try:
                subprocess.check_output(['python3', '--version']) 
                return 'python3'
            except:
                self.log(message='Python command not found, make sure python is installed!', error=True, discord=False)
                sys.exit(1)

    def log(self, message, icon="🤖", error=False, info=False, warning=False, chat=False, image_url="", console=True, discord=True):
        if not self.disable_logs:
            if self.use_return:
                logger.info(f"[{icon}] {message}")
            if console:
                if error:
                    logger.error(f"[{icon}] {message}")
                elif info:
                    logger.info(f"[{icon}] {message}")
                elif warning:
                    logger.warning(f"[{icon}] {message}")
                elif chat:
                    logger.info(f"[{icon}] {message}")
                else:
                    logger.info(f"[{icon}] {message}")
    
    def __wait_for_msa(self):
        @self.once('login')
        def await_login(*args):
            logger.info("Logged in successfully!")
            

    def __msa(self, *msa):
        self.msa_data = msa[0]
        self.msa_status = True
        self.log(message="It seems you are not logged in! Open your terminal for more information.", error=True,
                    console=False)
        msg = str(self.msa_data['message']).replace("\n", "")
        logger.error(f"It seems you are not logged in. {msg}")
        
        
        self.__wait_for_msa()
        if self.api_mode:
            self.bot.end()
            quit()
        self.msa_status = False

    def __versions_check(self):
        with self.console.status("[bold]Checking versions...\n"):
            # Node
            result = subprocess.run(["node", "--version"], 
                                    stdout=subprocess.PIPE,
                                    stderr=subprocess.PIPE,
                                    text=True)
            node_version = result.stdout.strip()
            # Remove leading 'v'
            node_version = node_version.removeprefix("v")
            # Remove periods
            node_version = node_version.replace('.', '')
            if int(node_version[:2]) >= 18:
                pass
            else:
                logger.warning(f"""
                                Detected node version {node_version[:2]} which isn't supported!
                                This may cause problems. Please update to node 18 or above!
                                """)

            # Pip
            pip_ = 'pip' if self.__check_python_command() == 'python' else 'pip3'
            result = subprocess.run([pip_, "--version"],
                                    stdout=subprocess.PIPE,
                                    stderr=subprocess.PIPE,
                                    text=True)
            pip_version = result.stdout.strip()
            # Remove
            match = re.search(r'pip (\d+(?:\.\d+)*).*python (\d+(?:\.\d+)*)', pip_version)
            if match:
                pip_version = match.group(1)
                python_version = match.group(2)
            return node_version, pip_version, python_version
    
    def __create_bot(self):
        if self.local_version == "auto" or self.local_version == "false":
            self.local_version = False
        else:
            self.version = str(self.local_version)
        
        if self.local_auth.lower() == "easymc":
            self.minecraft_protocol = require('minecraft-protocol')
            import types
            
            
            def easy_mc_auth(client, options):
                try:
                    res = requests.post('https://api.easymc.io/v1/token/redeem', headers={'Content-Type': 'application/json'}, json={"token":f"{options['easyMcToken']}"})
                    res_json =  res.json()
                    if res_json.get('error'):
                        raise Exception(f'EasyMC: {res_json["error"]}')
                        quit()
                    if not res_json:
                        raise Exception('Empty response from EasyMC.')
                        quit()
                    if len(res_json.get('session', '')) != 43 or len(res_json.get('mcName', '')) < 3 or len(res_json.get('uuid', '')) != 36:
                        raise Exception('Invalid response from EasyMC.')
                        quit()
                    session = {
                        'accessToken': res_json['session'],
                        'selectedProfile': {
                            'name': res_json['mcName'],
                            'id': res_json['uuid']
                        }
                    }
                    options.haveCredentials = True
                    client.session = session
                    options.username = client.username = session['selectedProfile']['name']
                    options.accessToken = session['accessToken']
                    client.emit('session', session)
                    options['connect'](client)
                except Exception as error:
                    print(error)
                    quit()


            try:
                local_bot = self.mineflayer.createBot({
                    'host': self.local_host,
                    'port': self.local_port,
                    'auth': easy_mc_auth,
                    'sessionServer': 'https://sessionserver.easymc.io',
                    'username': bytes(),
                    'version': self.local_version,
                    'checkTimeoutInterval': self.check_timeout_interval,
                    'logErrors': self.local_log_errors,
                    'hideErrors': self.local_hide_errors,
                    'keepAlive': self.local_keep_alive,
                })
                print(local_bot)
                self.bot = local_bot
                return local_bot
            except Exception as e:
                raise ValueError(f"Error while creating bot: {e}")
        else:    
            try:
                local_bot = self.mineflayer.createBot({
                    'host': self.local_host,
                    'port': self.local_port,
                    'username': self.local_username,
                    'password': self.local_password,
                    'auth': self.local_auth,
                    'version': self.local_version,
                    'onMsaCode': self.__msa,
                    'checkTimeoutInterval': self.check_timeout_interval,
                    'disableChatSigning': self.local_disable_chat_signing,
                    'profilesFolder': self.local_profiles_folder,
                    'logErrors': self.local_log_errors,
                    'hideErrors': self.local_hide_errors,
                    'keepAlive': self.local_keep_alive,
                    'loadInternalPlugins': self.local_load_internal_plugins,
                    'respawn': self.local_respawn,
                    'physicsEnabled': self.local_physics_enabled,
                    'defaultChatPatterns': self.local_default_chat_patterns
                })
                self.bot = local_bot
                return local_bot
            except Exception as e:
                raise ValueError(f"Error while creating bot: {e}")
            
        

    

    def __start(self):
        self.__setup_events()
        while not self.logged_in:
            time.sleep(1)
        self.log(
            f'Coordinates: {int(self.bot.entity.position.x)}, {int(self.bot.entity.position.y)}, {int(self.bot.entity.position.z)}',
            info=True)
        self.__load_plugins()

    def on(self, event: str):
        """
        Decorator for event registering

        ```python
        @bot.on('messagestr')
        def chat(_, message, *args):
            ...
        ```
        """
        def inner(function):
            On(self.proxy, event)(function)
        return inner
    
    def off(self, event: str, function):
        """
        Decorator for event unregistering

        ```python
        @bot.off('messagestr')
        def chat(_, message, *args):
            ...
        ```
        """
        print(function)
        off(emitter=self.proxy, event=f"{event}", handler=function)
        return
    

    def once(self, event: str):
        """
        Decorator for event registering

        ```python
        @bot.once('login')
        def login(*args):
            ...
        ```
        """
        def inner(function):
            Once(self.proxy, event)(function)
        return inner

    def emit(self, event: str, *params):
        """
        Emits an event which could be listened to

        ```python
        bot.emit('custom_chat', username, message)
        ```
        """
        self.bot.emit(event, *params)
        self.log(f"Emitting event {repr(event)}", info=True, discord=False)

    def add_method(self, target_name=None):
        """
        Decorator for adding a method dynamically.
        **IMPORTANT** You need the self argument! The self argument also points to your own plugin class, not the bot's

        ```python
        @add_method()
        def lobby(self, server="main"):
            self.bot.command("lobby", server)
        ```
        """
        def inner(function):
            def wrapper(*args, **kwargs):
                function(*args, **kwargs)
            wrapper.__name__ = function.__name__
            wrapper.__doc__ = function.__doc__

            if target_name:
                setattr(self, target_name, function)
            else:
                setattr(self, function.__name__, function)
            return wrapper
        return inner

    @cprop()
    def registry(self) -> Proxy: pass

    @cprop()
    def world(self) -> Proxy: pass

    @cprop()
    def entity(self) -> Proxy: pass

    @cprop()
    def entities(self) -> Proxy: pass

    @cprop()
    def username(self) -> str: pass

    @cprop()
    def spawn_point(self) -> Proxy: pass

    @cprop()
    def held_item(self) -> Proxy: pass

    @cprop()
    def using_held_item(self) -> bool: pass

    @property
    def game(self): return GameState(self.proxy.game)

    @property
    def creative(self): return CreativeMode(self.proxy.creative)
    
    @cprop()
    def physics_enabled(self) -> bool: pass
    
    @cprop()
    def player(self) -> Proxy: pass
    
    @cprop()
    def players(self) -> Proxy: pass
    
    @cprop()
    def tablist(self) -> Proxy: pass

    @cprop()
    def is_raining(self) -> bool: pass

    @cprop()
    def rain_state(self) -> int: pass

    @cprop()
    def thunder_state(self) -> int: pass

    @cprop()
    def chat_patterns(self) -> Proxy: pass

    @property
    def settings(self): return SettingsState(self.proxy.settings)

    @property
    def experience(self): return ExperienceState(self.proxy.experience)

    @cprop()
    def health(self) -> int: pass

    @cprop()
    def food(self) -> int: pass

    @cprop()
    def food_saturation(self) -> int: pass

    @cprop()
    def oxygen_level(self) -> int: pass

    @cprop()
    def physics(self) -> Proxy: pass

    @cprop()
    def firework_rocket_duration(self) -> int: pass

    @property
    def time(self): return TimeState(self.proxy.time)

    @cprop()
    def quick_bar_slot(self) -> int: pass

    @cprop()
    def inventory(self) -> Proxy: pass

    @cprop()
    def target_dig_block(self) -> Proxy: pass

    @cprop()
    def is_sleeping(self) -> bool: pass

    @cprop()
    def scoreboards(self) -> Proxy: pass

    @cprop()
    def scoreboard(self) -> Proxy: pass

    @cprop()
    def teams(self) -> Proxy: pass

    @cprop()
    def team_map(self) -> Proxy: pass

    @cprop()
    def control_state(self) -> Proxy: pass

    @cprop()
    def set_control_state(self) -> Callable[[str, bool], None]:
        """
        Returns Function(state: str, toggle: bool)
        """

    @cprop()
    def clear_control_states(self) -> Callable[[], None]:
        """
        Returns Function()
        """
        
    def __load_plugins(self):
        self.mc_data = require('minecraft-data')(self.bot.version)
        self.__pathfinder = require('mineflayer-pathfinder').pathfinder
        self.movements = require('mineflayer-pathfinder').Movements
        self.goals = require('mineflayer-pathfinder').goals
        
        self.bot.loadPlugin(self.__pathfinder)
        self.pathfinder = self.bot.pathfinder
        self.bot.loadPlugin(require('mineflayer-collectblock').plugin)
        self.movements = self.movements(self.bot, self.mc_data)
        self.bot.pathfinder.setMovements(self.movements)
    
    def __setup_events(self):
        @self.once("login")
        def on_login(*_):
            self.logged_in = True
            self.log(f"Connected to {self.local_host}", info=True)
            self.log(f'Logged in as {self.bot.username}', info=True)
            if not self.disable_viewer:
                self.__start_viewer()
            

        @self.on("path_update")
        def path_update(_, r):
            if not self.disable_viewer:
                path = [self.bot.entity.position.offset(0, 0.5, 0)]
                for node in r['path']:
                    path.append({'x': node['x'], 'y': node['y'] + 0.5, 'z': node['z']})
                self.bot.viewer.drawLine('path', path, 	0x0000FF)

        @self.on("death")
        def death(*_):
            self.log("Bot died" + ". Stopping bot!" * int(self.stop_bot_on_death), warning=True)
            if self.stop_bot_on_death:
                self.stop()
                quit()

        @self.on("kick")
        def kicked(_, reason, *__):
            self.log("Kicked from server" + ". Stopping bot!" * int(self.stop_bot_on_death) + f"\n\nReason: {reason}",
                     warning=True)
            if self.stop_bot_on_death:
                self.stop()
                quit()

        @self.on("error")
        def error(_, error):
            self.log(error, error=True)
            if error == "Authentication failed, timed out":
                self.stop()
                quit()

        @self.on("chat")
        def handleMsg(_, sender: str, message: str, *args):
            if self.enable_chat_logging:
                if not sender:
                    sender = "unknown"
                if not self.chat_database.contains(User.username == sender):
                    self.chat_database.insert({'username': sender, 'messages': [message]})
                else:
                    user = self.chat_database.get(User.username == sender)
                    existing_messages = user['messages']
                    existing_messages.extend([f"{message}"])
                    self.chat_database.update({'messages': existing_messages}, User.username == sender)
                self.log(f"{sender}: {message}", icon="💬", chat=True)

            if message.startswith(self.custom_command_prefix):
                command = message.removeprefix(self.custom_command_prefix)
                commanded = self.custom_commands.get(command, None)
                if commanded:
                    if sender == commanded["sender"]:
                        cmd, *params = command.split()
                        commanded["callback"](CommandContext(
                            sender, cmd, command, params, time.time(), args[1], self
                        ))

    def __start_viewer(self):
        try:
            self.mineflayer_viewer.mineflayer(self.bot, {"port": self.viewer_port})
            # _ = require("node-canvas-webgl")
            # @self.once("spawn")
            # def ___start_viewer(*args):
            #     self.mineflayer_viewer.headless(self.bot, { "output": f"127.0.0.1:{self.viewer_port}", "frames": 200, "width": 512, "height": 512 })
            self.log(f"Viewer started on port {self.viewer_port}", info=True)
        except Exception as e:
            print(e)
            self.log("There was an error while starting the viewer!", warning=True)
            
    
    def chat(self, *message):
        """
        Send a message in the chat
        """
        self.bot.chat(' '.join(message))
        
    def whisper(self, username, *message):
        """
        Send a whisper to a user with a message
        """
        self.bot.whisper(username, ' '.join(message))

    def command_safe(self, arg):
        if isinstance(arg, str) and arg.count(' '):
                return '"' + arg.replace('"', '\\"') + '"'
        elif isinstance(arg, bool):
            return repr(arg).lower()
        elif isinstance(arg, tuple):
            returns = []
            for i in arg:
                returns.append(self.command_safe(i))
            return returns
        else:
            return arg

    def command(self, command: str, *args):
        converted_args = []
        for arg in args:
            parsed = self.command_safe(arg)
            if isinstance(parsed, list):
                converted_args.extend(parsed)
            else:
                converted_args.append(parsed)
        self.chat('/' + command, *converted_args)


    def chat_history(self, username: str, server="") -> list:
        """
        Get all the chat history from a player in the database
        """
        if not self.enable_chat_logging:
            self.log(f"Chat logging is not enabled, set enableChatLogging=True in the bot config", warning=True)
            return []
        if server == "":
            server = self.local_host
        if os.path.exists(f"{server}".lower().replace(".", "") + "Database.json"):
            server_history = TinyDB(f"{server}".lower().replace(".", "") + "Database.json")
            user = server_history.get(User.username == username)
            if user:
                return user['messages']
            else:
                self.log(f"{username} has no chat history", warning=True)
        else:
            self.log(f"{server} has no database", warning=True)
            return []
        
    def clear_logs(self):
        """
        Clear all the chat logs from the database
        """
        if not self.enable_chat_logging:
            self.log(f"Chat logging is not enabled, set enableChatLogging=True in the bot config", warning=True)
            return
        self.chat_database.truncate()
        self.log("All databases are cleared!")
    
    def stop(self):
        """
        Stop the bot and all running code
        """
    
        self.bot.end()
        if not self.disable_viewer:
            self.bot.viewer.close()
        self.log("Ended bot!", warning=True)
        return
        
    def server_data(self, server:str=None) -> dict:
        """
        Get all the data from a server.
        if no server is provided the current server will be used
        """
        if server is None:
            server = self.local_host
        data = requests.get(f"https://api.mcstatus.io/v2/status/java/{server}").json()
        return data

    def set_data(self, item, value):
        """
        Sets custom data that can be later accessed. Also returns data.

        ```python
        bot.set_data("hello", "world")
        ... # some time consuming task later
        print(bot.get_data("hello")) # should print "world"
        ```
        """
        self.extra_data[item] = value
        return value

    def get_data(self, item, default: object = None, compare: object = "nothing to compare to"):
        """
        Gets custom data that is set prior. Also take in an optional compare parameter to do assertion with the obtained data.
        Default parameter for 'default' is None

        ```python
        # other code
        try:
            print(bot.get_data("custom_health", 200))
        except AssertionError:
            print("Bot is not at full health!")
        ```
        """
        result = self.extra_data.get(item, default)
        if not compare == "nothing to compare to": # there's a comparison
            if result != compare:
                raise AssertionError(
                    f"Incorrect value in custom data! Queried {repr(item)}={repr(result)}, instead expected {repr(item)}={repr(compare)}"
                )
        return result

    def set_prefix(self, new_prefix="!"):
        """
        Sets the bot's command prefix. This should not be empty as it will speed up responding to chat
        """
        self.custom_command_prefix = new_prefix

    def register_command(self, command_name: str, sender = None, returns: str | Callable[[CommandContext], None] = None, whisper: bool = True):
            """
            Registers a custom command that the bot listens to.

            Args:
                command_name (str): The name of the command to register.
                sender (Optional[str]): The sender of the command. If set, the bot will only respond if the senders match.
                returns (Optional[Union[str, Callable[[CommandContext], None]]]): The return value or callback function for the command.
                    - If a string is provided, the bot will respond with that string.
                    - If a callable function is provided, the bot will execute that function.
                whisper (bool): Indicates whether the bot should respond with a whisper. Default is True.

            Returns:
                None

            Examples:
                You can register a command with a return value using the following syntax:

                ```python
                bot.register_command("hello", returns="Hello there!")
                ```

                You can register a command with a callback function using the following syntax:

                ```python
                @bot.register_command("time")
                def get_time(ctx):
                    current_time = datetime.datetime.now()
                    ctx.respond(f"It is now {current_time}", whisper=True)
                ```

                Alternatively, you can use the `register_command` method as a decorator.

            Raises:
                TypeError: If the callback parameter is of an unsupported type.

            """
            command_sender = sender
            if isinstance(returns, str):
                self.custom_commands[command_name] = {
                    "sender": command_sender,
                    "callback": lambda ctx: ctx.respond(returns, whisper=whisper, whisper_to=command_sender)
                }
            elif callable(returns):
                self.custom_commands[command_name] = {
                    "sender": command_sender,
                    "callback": returns
                }
            elif returns is None:
                def inner(func):
                    def wrapper(ctx):
                        func(ctx)

                    wrapper.__name__ = func.__name__
                    wrapper.__doc__ = func.__doc__
                    wrapper.__dict__ = func.__dict__

                    self.custom_commands[command_name] = {
                        "sender": command_sender,
                        "callback": func
                    }
                    return wrapper
                return inner
            else:
                raise TypeError(
                    f"Cannot add custom command with callback of type {returns.__class__.__name__}!"
                )
            
    def collect_block(self, block:str, amount:int=1, max_distance:int=64):
        """
        Collect a block by name.

        Args:
            block (str): The name of the block to collect.
            amount (int, optional): The number of blocks to collect. Defaults to 1.
            max_distance (int, optional): The maximum distance to search for the block. Defaults to 64.

        Returns:
            None

        Raises:
            None

        Example:
            bot.collect_block("oak_log", amount=20)
        """
        # Get the correct block type
        self.collected = 0
        def collect():
            try:
                self.collected = 0 
                blockType = self.bot.registry.blocksByName[block]
                if not blockType:
                    return
                blocks = self.bot.findBlocks({
                    'matching': blockType.id,
                    'maxDistance': 64,
                    'count': amount
                })
                if len(blocks.valueOf()) == 0:
                    self.chat("I don't see that block nearby.")
                    return
                targets = []
                for i in range(min(len(blocks.valueOf()), amount)):
                    targets.append(self.bot.blockAt(blocks[i]))
                self.chat(f"Found {len(targets)}")
                try:
                    self.bot.collectBlock.collect(targets, timeout=10000)
                    # All blocks have been collected.
                    self.chat('Done')
                except Exception as err:
                    # An error occurred, report it.
                    print(err)
            except Exception as err:
                print(err)
        task = threading.Thread(target=collect)
        task.start()
        task.join()
        
        
    
    def goto(self, x: int, y: int, z: int, timeout: int = 600000000):
        """
        Go to the specified coordinates (x, y, z).

        Args:
            x (int): The x-coordinate.
            z (int): The z-coordinate.
            y (int, optional): The y-coordinate. Defaults to 0.
            timeout (int, optional): The timeout in milliseconds. Defaults to 600000000.

        Returns:
            None

        Examples:
            To move to the coordinates (10, 0, 20) with a timeout of 10 seconds:
            >>> bot.goto(10, 20, timeout=10000)
        """
        # Get the correct block type
        with self.console.status(f"[bold]Moving to ({x}, {y}, {z})...") as status:
            if y == None:
                self.bot.pathfinder.goto(self.goals.GoalNearXZ(int(x), int(z), 1), timeout=timeout)
            else:
                self.bot.pathfinder.goto(self.goals.GoalNear(int(x), int(y), int(z), 1), timeout=timeout)
            return
        
    def placeBlockWithOptions(self, referenceBlock, faceVector, options):
        dest = referenceBlock.position.plus(faceVector)
        oldBlock = self.bot.blockAt(dest)
        self.bot._genericPlace(referenceBlock, faceVector, options)
        newBlock = self.bot.blockAt(dest)
        if oldBlock.type == newBlock.type:
            @On(self.bot.world, f"blockUpdate:{dest}")
            def block_update(oldBlock, newBlock):
                print(newBlock)
                print(oldBlock)
                if not oldBlock or not newBlock or oldBlock.type != newBlock.type:
                    return
                else: 
                    raise Exception(f"No block has been placed: the block is still {oldBlock.name}")
        if not oldBlock and not newBlock:
            return
        if oldBlock and oldBlock.type == newBlock.type:
            raise Exception(f"No block has been placed: the block is still {oldBlock.name}")
        else:
            self.emit("blockPlaced", oldBlock, newBlock)

    def placeBlock(self, referenceBlock, faceVector, no_checks=False):
        if no_checks:
            try:
                dest = referenceBlock.position.plus(faceVector)
                # self.bot.lookAt(referenceBlock)
                oldBlock = self.bot.blockAt(dest)
                self.bot._genericPlace(referenceBlock, faceVector, { "swingArm": "right" })
                # self.bot.waitForTicks(ticks)
                
                newBlock = self.bot.blockAt(dest)
                self.emit("blockPlaced", oldBlock, newBlock)
                return True
            except:
                return False
        else:
            self.placeBlockWithOptions(referenceBlock, faceVector, { "swingArm": "right" })
            
def get_plugins():
    if os.path.isdir('plugins'):
        shutil.rmtree('plugins')
    git.Repo.clone_from("https://github.com/the-lodestone-project/Plugins", "plugins")
    try:
        source_dir = 'plugins/plugins'
        target_dir = 'plugins'
            
        file_names = os.listdir(source_dir)
            
        for file_name in file_names:
            shutil.move(os.path.join(source_dir, file_name), target_dir)
        
        shutil.rmtree('plugins/plugins')
    except:
        logger.warning("Plugins folder is empty!")
        pass
    
            

createBot = Bot



import structlog

structlog.configure(
    processors=[
        structlog.processors.TimeStamper(fmt="%H:%M:%S"),
        structlog.processors.add_log_level,
        structlog.dev.ConsoleRenderer(),
    ]
)

logger = structlog.getLogger(__name__)



import lodestone
import time
import asyncio
from queue import PriorityQueue
class Idkwhattocallthis:
    def __init__(self, bot: lodestone.bot):
        self.bot = bot
        self.__create_tree()
        self.__create_priority_queues()
        self.__add_tasks()
        self.__start()
        
    
    
    
    class create_task:
        def __init__(self, item_or_block, amount=1, obtain=None,craft=None):
            self.obtain = obtain
            self.craft = craft
            self.bot = Idkwhattocallthis.bot
            self.amount = amount
            self.item_or_block = item_or_block
            self.dependencies = []
            self.completed = False
            
        def add_dependency(self, task):
            self.dependencies.append(task)

        def is_available(self):
            return all(d.completed for d in self.dependencies)

        def get_next_task(self):
            return next(d for d in self.dependencies if not d.completed)
        
        def complete(self):
            print(f"Completing task: {self.item_or_block}")
            if self.action == "craft":
                try:
                    crafting_table = self.bot.bot.findBlocks({ 'matching': self.bot.bot.registry.blocksByName['crafting_table'].id, 'maxDistance': 20, 'count': 1 })[0]
                
                except:
                    print("error")
                    exit()
                self.bot.bot.pathfinder.goto(self.bot.pathfinder.goals.GoalNear(crafting_table.x, crafting_table.y, crafting_table.z, 1), timeout=600000000)
                print(self.craft_count)
                amount = int(self.craft_count)
                item = self.bot.bot.registry.itemsByName[self.item_or_block]
                craftingTableID = self.bot.bot.registry.blocksByName.crafting_table.id
                craftingTable = self.bot.bot.findBlock({
                    'matching': craftingTableID
                })
                if item:
                    recipe = self.bot.bot.recipesFor(item.id, None, 1, craftingTable)[0]
                    if recipe:
                        self.bot.bot.chat(f"I can make {self.item_or_block}")
                        try:
                            self.bot.bot.craft(recipe, amount, craftingTable)
                            self.bot.bot.chat(f"did the recipe for {self.item_or_block} {amount} times")
                        except Exception as err:
                            self.bot.bot.chat(f"error making {self.item_or_block}")
                    else:
                        self.bot.bot.chat(f"I cannot make {self.item_or_block}")
                else:
                    self.bot.bot.chat(f"unknown item: {self.item_or_block}")

            
            if self.obtain:
                self.bot.collect_block(f"{self.item_or_block}", amount=amount, max_distance=100)
            self.completed = True

        # Priority queues 

    class PriorityQueue:
        def __init__(self):
            self.queue = PriorityQueue()

        def put(self, priority, task):
            self.queue.put((priority, task))
        
        def get(self):
            return self.queue.get()[1]

        def empty(self):
            return self.queue.empty()

    
    def __create_tree(self):
        # Create sample tree

        self.wood = self.create_task(obtain=True, item_or_block="oak_log", bot=self.bot, amount=1)

        self.make_planks = self.create_task(action="craft", item_or_block="oak_planks", bot=self.bot)
        self.make_planks.add_dependency(self.chop_wood)

        self.make_stick = self.create_task(action="craft", item_or_block="stick", bot=self.bot)
        self.make_stick.add_dependency(self.make_planks)

        self.make_pickaxe = self.create_task(action="craft", item_or_block="wooden_pickaxe", bot=self.bot)
        self.make_pickaxe.add_dependency(self.make_planks)
        self.make_pickaxe.add_dependency(self.make_stick)

    
    def __create_priority_queues(self):
        # Create priority queues

        self.urgent_tasks = PriorityQueue()
        self.normal_tasks = PriorityQueue()


    def __add_tasks(self):
        # Add tasks 

        # urgent_tasks.put(10, defend_task)
        self.normal_tasks.put((5, self.make_pickaxe))


    def __process_tasks(self):
        if self.urgent_tasks.empty():
            task = self.normal_tasks.get()
        else:
            task = self.urgent_tasks.get()

        self.gather(task[1])

    # Gather function 

    def __gather(self, task: create_task):
        if task.is_available():
            task.complete()
            return

        dep = task.get_next_task()
        self.gather(dep)
        self.gather(task)

    # Main bot loop
    def __start(self):
        while not self.normal_tasks.empty() or not self.urgent_tasks.empty():
            self.__process_tasks()
        print("DONE!")


import requests
import structlog
from rich.console import Console
import asyncio
logger = structlog.get_logger()
console = Console()

def llm(input: str, data = ""):
    try:
        import g4f
    except:
        pass
    _providers = [
        g4f.Provider.Bing,
        g4f.Provider.GptGo,
        g4f.Provider.Bard
    ]

    output = []


    async def run_provider(provider: g4f.Provider.AsyncProvider):
    
        try:
            response = await provider.create_async(
                model=g4f.models.default.name,
                messages=[{"role": "user", "content": f"question about provided data: {input} data: {data} USE THIS DATA TO AWNSER THE QUESTION, KEEP IT SHORT"}],
            )
            output.append({provider.__name__: response})
        except Exception:
            logger.warning(f"[LLM] {provider.__name__} is not available.")

    async def run_all():
        calls = [
            run_provider(provider) for provider in _providers
        ]
        await asyncio.gather(*calls)
    with console.status(f"[bold green][LLM] Please wait...\n"):
        asyncio.run(run_all())
        
    default = g4f.ChatCompletion.create(
        model=g4f.models.gpt_4,
        messages=[{"role": "user", "content": f"question about provided data: {input} data: {data} USE THIS DATA TO AWNSER THE QUESTION, KEEP IT SHORT"}],
    )  # alterative model setting
    if "<!DOCTYPE html>" in default:
        logger.warning(f"[LLM] default is not available. Trying another...")
        defualtnew = g4f.ChatCompletion.create(
            model=g4f.models.gpt_35_turbo,
            messages=[{"role": "user", "content": f"question about provided data: {input} data: {data} USE THIS DATA TO AWNSER THE QUESTION, KEEP IT SHORT"}],
        )  # alterative model setting
        output.append({"base": defualtnew})
    else:
        output.append({"base": default})
    return output


def convert_case(string, case = "pascal"):
    match case:
        case "snake":
            new = []
            for seq in string.split("_"):
                new.append(seq.lower())
            name = "_".join(new)
        case "pascal":
            new = []
            for seq in string.split("_"):
                new.append(seq.title())
            name = "".join(new)
        case "camel":
            new = []
            one = True
            for seq in string.split("_"):
                new.append(seq.lower() if one else seq.title())
                one = False
            name = "".join(new)
        case _:
            name = string
    return name

def cprop(cap = "camel", proxy_name = ""):
    def decorator(func):
        @property
        def wrapped(self):
            name = proxy_name
            if not name:
                name = convert_case(func.__name__, cap)
            return getattr(self.proxy, name)
        try:
            wrapped.__name__ = func.__name__ 
        except AttributeError:
            pass # ignore if func is a property
        wrapped.__doc__ = func.__doc__
        try:
            wrapped.__annotations__ = func.__annotations__
        except AttributeError:
            pass # ignore if setting annotations fails

        return wrapped
    return decorator



from rich.console import Console
try:
    from discord import Embed
except:
    pass
import structlog
import datetime
import sys
import re
import time
from datetime import date
import os
import platform
import urllib.request
import json
import subprocess


logger = structlog.get_logger()

filestruc = "/"
if os.name == 'nt':
    filestruc = "\\"

__all__ = ['Server', 'createServer']

class Server:
    def __init__(
            self,
            motd: str = "Lodestone test server",
            *,
            maxPlayers: int = 20,
            installPath: str = "LodestoneServer",
            serverType: str = "Paper",
            ram: int = 2048,
            onlineMode: bool = True,
            logging: bool = True,
            gameMode: int = 1,
            difficulty: int = 1,
            worldFolder: str = "world",
            generation: dict = {'name': 'diamond_square', 'options': {'worldHeight': 80}},
            kickTimeout: int = 10000,
            modpe: bool = False,
            viewDistance: int = 10,
            playerListText: dict = {'header': 'Flying squid','footer': 'Test server'},
            everybodyOp: bool = True,
            maxEntities: int = 100,
            version: str = "1.20",
            port: int = 25565,
            

            ls_disable_logs: bool = False,
            ls_skip_checks: bool = False,
            ls_debug_mode: bool = False,
            ls_discord_webhook: str = None,
            ls_use_discord_forums: bool = False,
    ):
        """
        Create the bot. Parameters in camelCase are passed into mineflayer. Parameters starting with ls_ is Lodestone specific
        """
        if ls_debug_mode:
            os.environ["DEBUG"] = "minecraft-protocol"
        else:
            os.environ["DEBUG"] = ""

        self.local_motd = motd
        self.local_ram = ram
        self.local_max_players = maxPlayers
        self.local_port = port
        self.local_server_type = serverType
        self.local_version = version
        self.local_install_path = installPath
        self.local_online_mode = onlineMode
        self.local_logging = logging
        self.local_game_mode = gameMode
        self.local_difficulty = difficulty
        self.local_world_folder = worldFolder
        self.local_generation = generation
        self.local_kick_timeout = kickTimeout
        self.local_modpe = modpe
        self.local_view_distance = viewDistance
        self.local_player_list_text = playerListText
        self.local_everybody_op = everybodyOp

        self.local_max_entities = maxEntities
        self.disable_logs = ls_disable_logs
        self.skip_checks = ls_skip_checks
        self.discord_webhook = ls_discord_webhook
        self.use_discord_forums = ls_use_discord_forums

        self.console = Console()
        self.extra_data = {}
        self.loaded_plugins = {}

        if not self.skip_checks:
            self.node_version, self.pip_version, self.python_version = self.__versions_check()
        else:
            self.node_version, self.pip_version, self.python_version = "unknown", "unknown", "unknown"

        # if self.discord_webhook is not None:
        #     embed = Embed(
        #         title="Successfully Connected to Webhook!",
        #         description=f"""
        #         **Great news!** The bot has successfully connected to this channel's webhook.
        #         From now on, it will send all the logs and valuable data right here, keeping you informed about everything happening on the server.
                
        #         **Versions: **
        #         * [**Node**](https://nodejs.org/): {self.node_version}
        #         * [**Pip**](https://pypi.org/project/pip/): {self.pip_version}
        #         * [**Python**](https://www.python.org/): {self.python_version}
                
        #         **Links: **
        #         * [**GitHub**](https://github.com/SilkePilon/Lodestone)
        #         * [**Report Bugs**](https://github.com/SilkePilon/Lodestone/issues)
        #         * [**Web Interface**](https://github.com/SilkePilon/Mineflayer.py-react)
        #         """,
        #         color=0x3498db
        #     )
        #     embed.timestamp = datetime.datetime.utcnow()
        #     embed.set_footer(text='\u200b', icon_url="https://github.com/SilkePilon/Lodestone/blob/main/chestlogo.png?raw=true")
        #     if ls_use_discord_forums:
        #         today = date.today()
        #         send_webhook(ls_discord_webhook, content=f"{today}", thread_name=f"{today}", username="Lodestone", avatar_url="https://github.com/SilkePilon/Lodestone/blob/main/chestlogo.png?raw=true", embed=embed)
        #     else:
        #         try:
        #             send_webhook(ls_discord_webhook, content="", username="Lodestone", avatar_url="https://github.com/SilkePilon/Lodestone/blob/main/chestlogo.png?raw=true", embed=embed)
        #         except Exception as e:
        #             print(e)
        #             logger.error(f"Detected that you are using a Forums channel but 'useDiscordForums' is set to False. Please change 'useDiscordForums' to True or provide a webhook url for a text channel.")

        self.python_command = self.__check_python_command()
        if not ls_skip_checks:
            with self.console.status("[bold green]Checking for updates...\n") as status:
                status.update("[bold green]Updating pip package...\n")
                os.system(f'{self.python_command} -m pip install -U lodestone >/dev/null 2>&1')
        self.script_directory = os.path.dirname(os.path.abspath(sys.argv[0]))
        self.__create_server()
    
    def __check_python_command(self):
        try:
            subprocess.check_output(['python', '--version'])
            return 'python'
        except:
            try:
                subprocess.check_output(['python3', '--version']) 
                return 'python3'
            except:
                self.__logging(message='Python command not found, make sure python is installed!', error=True,
                               discord=False)
                sys.exit(1)

    def __logging(self, message, icon="💾", error=False, info=False, warning=False, chat=False, image_url="", console=True, discord=True):
        if not self.disable_logs:
            if self.discord_webhook and discord:
                try:
                    from discord import Embed
                except:
                    pass
                color = 0x3498db
                if error:
                    color = 0x992d22
                elif info:
                    color = 0x3498db
                elif warning:
                    color = 0xe67e22
                elif chat:
                    color = 0x2ecc71
                embed = Embed(title="", description=f"**[{icon}] {message}**", color=color)
                embed.timestamp = datetime.datetime.utcnow()
                if image_url != "":
                    embed.set_thumbnail(url=image_url)
                try:
                    embed.set_footer(text=f'{self.bot.username}', icon_url=f"https://mc-heads.net/avatar/{self.bot.username}/600.png")
                except:
                    embed.set_footer(text='\u200b', icon_url="https://github.com/SilkePilon/Lodestone/blob/main/chestlogo.png?raw=true")
                # if self.use_discord_forums:
                #     today = date.today()
                #     send_webhook(self.discord_webhook, content=f"{today}", thread_name=f"{today}", username="Lodestone", avatar_url="https://github.com/SilkePilon/Lodestone/blob/main/chestlogo.png?raw=true", embed=embed)
                # else:
                #     try:
                #         send_webhook(self.discord_webhook, content=f"", username="Lodestone", avatar_url="https://github.com/SilkePilon/Lodestone/blob/main/chestlogo.png?raw=true", embed=embed)
                #     except Exception as e:
                #         print(e)
                #         logger.error(f"Detected that you are using a Forums channel but 'useDiscordForums' is set to False. Please change 'useDiscordForums' to True or provide a webhook url for a text channel.")
            if console:
                if error:
                    logger.error(f"[{icon}] {message}")
                elif info:
                    logger.info(f"[{icon}] {message}")
                elif warning:
                    logger.warning(f"[{icon}] {message}")
                elif chat:
                    logger.info(f"[{icon}] {message}")
                else:
                    logger.info(f"[{icon}] {message}")

    def __versions_check(self):
        with self.console.status("[bold green]Checking versions..."):
            # Node
            result = subprocess.run(["node", "--version"], 
                                    stdout=subprocess.PIPE,
                                    stderr=subprocess.PIPE,
                                    text=True)
            node_version = result.stdout.strip()
            # Remove leading 'v'
            node_version = node_version.removeprefix("v")
            # Remove periods
            node_version = node_version.replace('.', '')
            if int(node_version[:2]) >= 18:
                pass
            else:
                logger.warning(f"""
                                Detected node version {node_version[:2]} which isn't supported!
                                This may cause problems. Please update to node 18 or above!
                                """)

            # Pip
            pip_ = 'pip' if self.__check_python_command() == 'python' else 'pip3'
            result = subprocess.run([pip_, "--version"],
                                    stdout=subprocess.PIPE,
                                    stderr=subprocess.PIPE,
                                    text=True)
            pip_version = result.stdout.strip()
            # Remove
            match = re.search(r'pip (\d+(?:\.\d+)*).*python (\d+(?:\.\d+)*)', pip_version)
            if match:
                pip_version = match.group(1)
                python_version = match.group(2)
            return node_version, pip_version, python_version

    def __create_server(self):


      # SETUP PROGRESS BAR
      global pbar
      pbar = None
      console = Console()
      status = console.status("[bold green]Downloading...")
      


      def get_folder():
        path = str(self.local_install_path)
        isExist = os.path.exists(path + "/")
        if not isExist:
          os.makedirs(path)
        if path is not None:
            # Getting the list of directories
            dir = os.listdir(path)
            
            # Checking if the list is empty or not
            if len(dir) == 0:
                # Dir is empty
                return path
            else:
                self.__logging("Directory isn't empty", error=True)
                exit()

      def get_server_type():
        server_type = str(self.local_server_type).lower()
        if (server_type == "paper") or (server_type == "vanilla"):
            return server_type.capitalize()
        else:
            self.__logging("Enter a valid answer.", error=True)
            exit()


      def get_latest_paper(version):
          self.__logging("Getting latest paper server url from: https://papermc.io", info=True)

          with urllib.request.urlopen(f"https://papermc.io/api/v2/projects/paper/versions/{version}") as baseurl:
              basedata = json.loads(baseurl.read().decode())
              buildsList = []
              for i in basedata["builds"]:
                  buildsList.append(i)
              
              self.__logging(f"Found correct version, {version}: searching for download link...", info=True)
              
              with urllib.request.urlopen(f"https://papermc.io/api/v2/projects/paper/versions/{version}/builds/{buildsList[-1]}") as buildurl:
                  builddata = json.loads(buildurl.read().decode())
                  name = builddata["downloads"]["application"]["name"]
                  final = f"https://papermc.io/api/v2/projects/paper/versions/{version}/builds/{buildsList[-1]}/downloads/{name}"
                  return final

      def download_file(path, url, name):
          self.__logging(f"Downloading server to {path}, as {name}", info=True)
          self.__logging("Downloading latest server file: ", info=True)
          with console.status("[bold green]Downloading...") as status:
            urllib.request.urlretrieve(url, os.path.join(path, name))
            status.update("[bold green]Done...")
          self.__logging("Download Complete!", info=True)


      def get_latest_vanilla(version):
          self.__logging("Getting latest vanilla server url from: https://launchermeta.mojang.com/mc/game/version_manifest.json", info=True)
          
          # Get json file with all versions
          with urllib.request.urlopen("https://launchermeta.mojang.com/mc/game/version_manifest.json") as url:
              data = json.loads(url.read().decode())
              for i in data["versions"]:
                  id = i["id"]
                  versionJson = i["url"]
                  
                  if id == version:
                      self.__logging(f"Found correct version, {version}: searching for download link...", info=True)

                      with urllib.request.urlopen(versionJson) as versionUrl:
                          dataUrl = json.loads(versionUrl.read().decode())
                          download = dataUrl["downloads"]["server"]["url"]

                          return download                   

      def first_run(path):
            self.__logging("The server will now start and stop to create folders.", info=True)
            time.sleep(1)
            os.chdir(os.path.abspath(path)) # CD into server folder
            try:
                result = subprocess.run(['java', '-jar', 'server.jar'], check = True)
            except subprocess.CalledProcessError:
                self.__logging("Could not start the server, make sure Java is installed.", error=True)
                exit()

            
            try:
              with open('eula.txt', 'r+') as file:
                  for line in file:
                      print(line.replace('eula=false', 'eula=true'), end='')
              self.__logging("Eula accepted.", info=True)
            except FileNotFoundError:
              self.__logging("Unsupported Java detected!", error=True)
              exit()

      def modify_props():
        serv_prop = open("server.properties", "r+")
        lines_prop = serv_prop.readlines()
        
        # seed = str(input("\nWhich seed would you like to set: "))
        # lines_prop[4] = f"level-seed={seed}\n"
        
        # self.__logging(f"Seed changed to {seed}.", info=True)

        name = str(self.local_motd)
        lines_prop[10] = f"motd={name}\n"

        self.__logging(f"Server name changed to {name}.", info=True)


        players = str(self.local_max_players)
        lines_prop[18] = f"max-players={players}\n"
        
        self.__logging(f"Player limit changed to {players}.", info=True)


        view = str(self.local_view_distance)
        lines_prop[23] = "view-distance=" + view + "\n"
        
        self.__logging(f"Render distance changed to {view}.", info=True)


        port = str(self.local_port)
        lines_prop[27] = f"server-port={port}\n"
        
        self.__logging(f"Server port changed to {port}.", info=True)
        

        online = self.local_online_mode
        if online == True:
            online_check = "enabled."
            lines_prop[19] = "online-mode=true\n"
            
        if online == False:
            online_check = "disabled."
            lines_prop[19] = "online-mode=false\n"
            
        self.__logging(f"Online mode {online_check}", info=True)


        self.__logging("Saving config and exiting...", info=True)

        # Save file
        serv_prop.writelines(lines_prop)
        serv_prop.close()

      def make_start_script():
          running = True
          while running:

             
                  # Finds running OS
                  running_os = platform.system()

                 
                  # Linux/macOS - make .sh file
                  if running_os == "Linux" or os == "Darwin":
                      if self.local_server_type.lower() == "paper":
                        script_type = "y"
                      else:
                        script_type = "n"
                      start_sh = open("start.sh", "x")

                      # Optimized start script
                      if script_type == "y":
                          ram = str(self.local_ram)
                          
                          start_sh.writelines('#!/bin/bash\njava -Xmx'+ram+'M -Xms'+ram+'M -XX:+UseG1GC -XX:+ParallelRefProcEnabled -XX:MaxGCPauseMillis=200 -XX:+UnlockExperimentalVMOptions -XX:+DisableExplicitGC -XX:+AlwaysPreTouch -XX:G1NewSizePercent=30 -XX:G1MaxNewSizePercent=40 -XX:G1HeapRegionSize=8M -XX:G1ReservePercent=20 -XX:G1HeapWastePercent=5 -XX:G1MixedGCCountTarget=4 -XX:InitiatingHeapOccupancyPercent=15 -XX:G1MixedGCLiveThresholdPercent=90 -XX:G1RSetUpdatingPauseTimePercent=5 -XX:SurvivorRatio=32 -XX:+PerfDisableSharedMem -XX:MaxTenuringThreshold=1 -Dusing.aikars.flags=https://mcflags.emc.gs -Daikars.new.flags=true -jar server.jar nogui')
                          start_sh.close()
                          # Make file executable
                          os.chmod('start.sh', 0o777)
                          running2 = False
                          running = False

                          self.__logging("Created script", info=True)

                      # Normal start script
                      elif script_type == "n":
                          ram = str(self.local_ram)
                          
                          start_sh.writelines('#!/bin/bash\njava -Xmx'+ram+'M -Xms'+ram+'M -jar server.jar nogui')
                          start_sh.close()
                          # Make file executable
                          os.chmod('start.sh', 0o777)
                          running2 = False
                          running = False

                          self.__logging("Created script", info=True)

                      # Wrong answer
                      else:
                          self.__logging("Enter a valid answer.", warning=True)

                  # Windows - make .bat file
                  elif running_os == "Windows":
                      if self.local_server_type.lower() == "paper":
                        script_type = "y"
                      else:
                        script_type = "n"
                      start_sh = open("start.bat", "x")

                      # Optimized start script
                      if script_type == "y":
                          ram = str(self.local_ram)
                          
                          start_sh.writelines('java -Xmx'+ram+'M -Xms'+ram+'M -XX:+UseG1GC -XX:+ParallelRefProcEnabled -XX:MaxGCPauseMillis=200 -XX:+UnlockExperimentalVMOptions -XX:+DisableExplicitGC -XX:+AlwaysPreTouch -XX:G1NewSizePercent=30 -XX:G1MaxNewSizePercent=40 -XX:G1HeapRegionSize=8M -XX:G1ReservePercent=20 -XX:G1HeapWastePercent=5 -XX:G1MixedGCCountTarget=4 -XX:InitiatingHeapOccupancyPercent=15 -XX:G1MixedGCLiveThresholdPercent=90 -XX:G1RSetUpdatingPauseTimePercent=5 -XX:SurvivorRatio=32 -XX:+PerfDisableSharedMem -XX:MaxTenuringThreshold=1 -Dusing.aikars.flags=https://mcflags.emc.gs -Daikars.new.flags=true -jar server.jar nogui')
                          start_sh.close()

                          running2 = False
                          running = False

                          self.__logging("Created script", info=True)

                      # Normal start script
                      elif script_type == "n":
                          ram = str(self.local_ram)
                          
                          start_sh.writelines('java -Xmx'+ram+'M -Xms'+ram+'M -jar server.jar nogui')
                          start_sh.close()

                          running2 = False
                          running = False

                          self.__logging("Created script", info=True)
                  


      def main():

          running = True
          while running:
              
              # Ask where to install server
              path = get_folder()

              # Ask server type
              if get_server_type() == "Vanilla":
                  # Tries to find vanilla URL
                  get_dl_question = True
                  while get_dl_question:
                      version = str(self.local_version)
                      url = get_latest_vanilla(version)
                      if url is None:
                          self.__logging(f"Didn't find a download link for version {version}.", error=True)
                          continue
                      else:
                          self.__logging(f"Found download link for version {version}: {url}", info=True)

                          # Download
                          download_file(path, url, "server.jar")

                          get_dl_question = False # Go to next question
              else:
                  #Tries to find paper URL
                  get_dl_question = True
                  while get_dl_question:
                      version = str(self.local_version)
                      url = get_latest_paper(version)
                      if url is None:
                          self.__logging(f"Didn't find a download link for version {version}.", error=True)
                          continue
                      else:
                          self.__logging(f"Found download link for version {version}: {url}", info=True)

                          # Download
                          download_file(path, url, "server.jar")

                          get_dl_question = False # Go to next question

              # Setup server
              first_run(path)
              modify_props()

              # Finish server + start script
              make_start_script()

              self.__logging("Server is done, simply start it by running the start script!", info=True)
              running = False


      main()

      
      
          

createServer = Server


import os
import sys
import click
from lodestone import fastapi
import subprocess
import time
from rich.console import Console
import lodestone
from dotenv import load_dotenv
import requests

@click.command(context_settings={"ignore_unknown_options": True})
@click.option("--console", "-c", default=False, is_flag=True, help="Force the app to use the console")
@click.option("--test", "-t", is_flag=True, help="Run version tests on Lodestone")
@click.option("--host", "-h", help="Host to run Lodestone's gui on", default=None)
@click.option("--port", "-p", help="Port to run Lodestone's gui on", default=8000)
@click.argument('args', nargs=-1)
def run(console, test, host, port, args):
    
    if test:
        versions = ["auto", "1.20.1", "1.19", "1.18", "1.17", "1.16", "1.15", "1.14", "1.13", "1.11", "1.10", "1.9", "1.8"]

        console = Console()

        with console.status("[bold]Checking versions...\n") as status:
            for version in versions:
                failed = ""
                status.update(f"Checking version {version}")

                # Create bot with the current version  
                try:
                    bot = lodestone.createBot(
                        host='2b2t.org',
                        username='TestBot',  
                        version=version,
                        checkTimeoutInterval=70,
                        ls_skip_checks=True,
                        hideErrors=False,
                        ls_debug_mode=True,
                        ls_disable_viewer=True
                    )
                except:
                    failed = "failed ([bold red]❌[reset])"
                    print(f"Version {version} failed to connect")
                    
                
                # Check connection
                if bot.tablist:
                    failed = "passed ([bold green]✔️[reset])"
                    status.update(f"Checking version {version} {failed}")
                else:
                    failed = "failed ([bold red]❌[reset])"
                    print(f"Version {version} failed to connect")
                
                # Stop bot
                bot.stop()

                # Add delay to avoid congestion
                for i in range(0, 10):
                    
                    status.update(f"{version} {failed} ({10 - i} until next version)")
                    time.sleep(1)
    
    def check_python_command():
        try:
            subprocess.check_output(['python', '--version'])
            return 'python'
        except:
            try:
                subprocess.check_output(['python3', '--version']) 
                return 'python3'
            except:
                print('Python command not found')
                sys.exit(1)

    python_command = check_python_command()

    is_linux = sys.platform.startswith('linux')
    is_mac = sys.platform == 'darwin'
    is_windows = sys.platform in ['win32', 'cygwin']

    script_directory = os.path.dirname(os.path.abspath(__file__))

    if console:
        os.system(f"{python_command} {script_directory}/console.py {' '.join(args)}")
    else:
        if host is None:
            import socket
            s = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
            s.connect(("8.8.8.8", 80))
            host = s.getsockname()[0]
            s.close()
            
        load_dotenv()
    
        username = os.getenv("LODESTONE_USERNAME")
        password = os.getenv("LODESTONE_PASSWORD")
        if username is None:
            print("For security reasons, you must set a username and password to protect your account.")
            username = input("Username: ")
            password = input("Password: ")
            os.environ["LODESTONE_USERNAME"] = f"{username}"
            os.environ["LODESTONE_PASSWORD"] = f"{password}"
            with open(".env", "w") as f:
                f.write(f"LODESTONE_USERNAME={username}\nLODESTONE_PASSWORD={password}")
                print("Login details saved to .env file. You can change these at any time.")
        
        if not os.path.isfile("favicon.ico"):
            img_data = requests.get("https://github.com/the-lodestone-project/Lodestone/raw/main/assets/favicon.png").content
            with open('favicon.png', 'wb') as handler:
                handler.write(img_data)
        
        
        try:
            lodestone.ui.queue().launch(server_name=f"{host}", server_port=port, show_api=False, auth=(f'{os.environ["LODESTONE_USERNAME"]}', f'{os.environ["LODESTONE_PASSWORD"]}'), share=False, quiet=True, favicon_path="favicon.png", auth_message="Please login with your set username and password. These are not your Minecraft credentials.")
        except OSError:
            raise OSError(f"Port {port} is already in use!")

if __name__ == "__main__":
    run()


import os

if os.path.isdir('plugins'):
    try:
        from lodestone.plugins import Base as plugins
    except:
        from .plugins import Base as plugins
else:
    import lodestone
    lodestone.bot.get_plugins()
    try:
        from lodestone.plugins import Base as plugins
    except:
        from .plugins import Base as plugins

try:
    from lodestone.bot import createBot, Bot
    from lodestone.utils import llm
    from lodestone.api import fastapi
    from lodestone.server import createServer, Server
    from javascript import On as Event
    from lodestone.logger import logger
    from lodestone.gradios import ui
except:
    from .bot import createBot, Bot
    from .utils import llm
    from .api import fastapi
    from .server import createServer, Server
    from javascript import On as Event
    from .logger import logger
    from .gradios import ui
    



import gradio as gr
import lodestone
from lodestone import plugins
import time

global created
created = False
global chat_history
chat_history = []
global plugin_list
plugin_list = []


css = """
.error {
  color: red
}
"""



def get_bot_status():
    if 'bot' in globals():
        return "Stop Bot"
    else:
        return "Start/Stop Bot"

def change_tab():
    return gr.Tabs.update(selected=1)



def create(email, auth, host, port, version, viewer, plugin, enable_viewer, skip_checks):
    try:
        if 'bot' in globals():
            def stop_bot():
                try:
                    global bot
                    bot.stop()
                    gr.Info("Successfully stopped bot!")
                    del bot
                except Exception as e:
                    print(e)
                    pass
            stop_bot()
            return "Already logged in!", "Already logged in!", "Login/Create Bot"
    except Exception as e:
        print(e)
        pass
    if not host or not email or not version:
        gr.Warning("not all fields are filled in!")
        return "Unknown", "Unknown", "Login/Create Bot"
    
    
    global bot
    plugin_str = ""
    if plugin:
        plugin_str += "Plugins: "
        for new_plugin in plugin:
            if new_plugin == "Schematic Builder":
                plugin_str += "Schematic Builder, "
                plugin_list.append(plugins.schematic)
            elif new_plugin == "Cactus Farm Builder":
                plugin_str += "Cactus Farm Builder, "
                plugin_list.append(plugins.cactus)
            elif new_plugin == "Discord Rich Presence":
                plugin_str += "Discord Rich Presence, "
                plugin_list.append(plugins.discordrp)
            else:
                pass
    if enable_viewer == False:
        enable_viewer = True
    elif enable_viewer == True:
        enable_viewer = False
    bot = lodestone.createBot(
        host=host,
        username=email,
        port=port,
        ls_viewer_port=viewer,
        version=version,
        profilesFolder="./cache",
        auth=auth,
        ls_disable_viewer=enable_viewer,
        ls_skip_checks=skip_checks,
        ls_plugin_list=plugin_list
    )
    
    @bot.on('messagestr')
    def chat_history_add(this, message, messagePosition, jsonMsg, sender, *args):
        message = str(message).replace("\n","")
        if str(sender).lower() == "none":
            chat_history.append(f"{message}")
        else:
            chat_history.append(f"{sender}: {message}")
    
    info =f"""Successfully logged in as {bot.username}"""
    
    
    
    gr.Info(f"Successfully logged in as {bot.username}\n{plugin_str}")
    return bot.username, info, "Stop Bot"


def create_multiple(email, auth, host, port, version, amount):
    try:
        if 'bot' in globals():
            def stop_bot():
                try:
                    global bot
                    bot.stop()
                    gr.Info("Successfully stopped bot!")
                    del bot
                except Exception as e:
                    print(e)
                    pass
            stop_bot()
            return "Already logged in!", "Already logged in!", "Login/Create Bot"
    except Exception as e:
        print(e)
        pass
    if not host or not email or not version:
        gr.Warning("not all fields are filled in!")
        return "Unknown", "Unknown", "Login/Create Bot"
    
    
    for bots in range(0, amount):
        lodestone.createBot(
            host=host,
            username=email+str(bots),
            port=port,
            ls_disable_viewer=True,
            version=version,
            profilesFolder="./cache",
            auth=auth,
            ls_skip_checks=True,
        )
    
    global bot
    bot = amount
    # @bot.on('messagestr')
    # def chat(this, message, messagePosition, jsonMsg, sender, *args):
    #     message = str(message).replace("\n","")
    #     if str(sender).lower() == "none":
    #         chat_history.append(f"{message}")
    #     else:
    #         chat_history.append(f"{sender}: {message}")
    
    info =f"""Successfully created {amount} bots!"""
    
    
    
    gr.Info(f"Successfully created {amount} bots!")
    return amount, info, "Stop Bot"



def get_username():
    if 'bot' in globals():
        return bot.username
    else:
        return "None"



def get_player_health():
    if 'bot' in globals():
        return bot.health
    else:
        return "Unknown"

def get_player_food():
    if 'bot' in globals():
        return bot.food
    else:
        return "Unknown"
    
def get_player_experience():
    if 'bot' in globals():
        return bot.experience.level
    else:
        return "Unknown"
    
def get_player_difficulty():
    if 'bot' in globals():
        
        if bot.settings.difficulty == 0:
            return "peaceful"
        elif bot.settings.difficulty == 1:
            return "easy"
        elif bot.settings.difficulty == 2:
            return "normal"
        elif bot.settings.difficulty == 3:
            return "hard"
    
    else:
        return "Unknown"
    

def get_all_data():
    if 'bot' in globals():
        return bot.player
    else:
        return "Unknown"




def get_latest_chats():
    if 'bot' in globals():
        if len(chat_history) > 30:
            chat_history.clear()
            return "No chat messages yet!"
        string = ""
        for i in chat_history[-9:]:
            string += i + "\n"
        return string
    else:
        return "No chat messages yet!"



# def upload_file(files):
#     global build_file
#     file_paths = [file.name for file in files]
#     build_file = file_paths
#     print(file_paths)
#     return file_paths


def build_schematic(files, x, z):
    if not x or not z or not files:
        gr.Warning("not all fields are filled in!")
        return
    if 'bot' in globals():
        bot.goto(x, z)
        time.sleep(2)
        bot.build_schematic(f'{files.name}')
    else:
        gr.Warning("You need to login first!")
        


with gr.Blocks(theme=gr.themes.Soft(), title="The Lodestone Project") as ui:
    with gr.Tab("Bot Settings"):
        # gr.Markdown(requests.get('https://raw.githubusercontent.com/the-lodestone-project/Lodestone/main/README.md').text)
        # gr.Image("https://github.com/the-lodestone-project/Lodestone/blob/main/assets/logo.png?raw=true", min_width=2000)
        with gr.Tab("Signle Bot"):
            email = gr.Textbox(placeholder="Notch", label="Username",info="Username to login with")
            auth = gr.Dropdown(["microsoft", "offline"], value="microsoft", label="Authentication Method",info="Authentication method to login with")
            host = gr.Textbox(placeholder="2b2t.org", label="Server Ip",info="Server ip to connect to")
            port = gr.Number(value=25565, label="Sever Port", info="Server port to connect to. Most servers use 25565",precision=0)
            version = gr.Dropdown(["auto","1.20", "1.19", "1.18", "1.17", "1.16.4", "1.16", "1.15", "1.14", "1.13", "1.12", "1.11", "1.10", "1.9", "1.8"], value="auto", label="Version",info="Version to connect with. Use auto to automatically detect the version of the server")
            with gr.Accordion("Optional Settings", open=False):
                enable_viewer = gr.Checkbox(value=True, label="Enable Viewer", info="Enable the viewer to see the bot's view",interactive=True)
                skip_checks = gr.Checkbox(value=True, label="Skip Checks/Updates", info="Skip checks to speed up the bot",interactive=True)
                viewer = gr.Number(value=5001, label="Viewer Port", info="Viewer port to display the bot's view",precision=0)
                plugin = gr.Dropdown(["Schematic Builder", "Cactus Farm Builder", "Discord Rich Presence"],multiselect=True, label="Plugins",info="Plugins to load on startup")
            btn = gr.Button(value=get_bot_status,variant='primary')
            
            
            
            
            out_username = gr.Textbox(value="", label="Logged in as")
            info = gr.Textbox(value="", label="Info")
            
            btn.click(create, inputs=[email, auth, host, port, version, viewer, plugin, enable_viewer, skip_checks], outputs=[out_username, info, btn], show_progress="minimal")
        # with gr.Tab("Parameters"):
        #     def live_view():
        #         try:
        #             if bot.viewer_port:
        #                 port = bot.viewer_port
        #             else:
        #                 port = 5001
        #         except:
        #             port = 5001
        #         return f'<iframe src="http://localhost:{port}" style="width:50vw; height:50vh;">Your browser doesnt support iframes</iframe>'
        #     gr.HTML(f'<iframe src="http://localhost:{port}" style="width:50vw; height:50vh;">Your browser doesnt support iframes</iframe>')
        #     pass
        with gr.Tab("Multiple Bot"):
            email = gr.Textbox(placeholder="Notch", label="Username Prefix",info="Username prefix. The bot will login with this prefix and a number after it")
            auth = gr.Dropdown(["offline"], value="offline", label="Authentication Method",info="Authentication method to login with")
            host = gr.Textbox(placeholder="2b2t.org", label="Server Ip",info="Server ip to connect to")
            port = gr.Number(value=25565, label="Sever Port", info="Server port to connect to. Most servers use 25565",precision=0)
            version = gr.Dropdown(["auto","1.20", "1.19", "1.18", "1.17", "1.16.4", "1.16", "1.15", "1.14", "1.13", "1.12", "1.11", "1.10", "1.9", "1.8"], value="auto", label="Version",info="Version to connect with. Use auto to automatically detect the version of the server")
            amount = gr.Slider(minimum=1, maximum=50, step=1, label="Amount", info="Amount of bots to create", interactive=True)
            with gr.Accordion("Optional Settings", open=False):
                enable_viewer = gr.Checkbox(value=False, label="Enable Viewer", info="Enable the viewer to see the bot's view",interactive=False)
                viewer = gr.Number(value=5001, label="Viewer Port", info="Viewer port to display the bot's view",precision=0, interactive=False)
                # plugin = gr.Dropdown(["Schematic Builder", "Cactus Farm Builder", "Discord Rich Presence"],multiselect=True, label="Plugins",info="Plugins to load on startup")
            btn = gr.Button(value=get_bot_status,variant='primary')
            
            
            
            
            out_username = gr.Textbox(value="", label="Bot count")
            info = gr.Textbox(value="", label="Info")
            
            btn.click(create_multiple, inputs=[email, auth, host, port, version, amount], outputs=[out_username, info, btn], show_progress="minimal")
        
    with gr.Tab("Chat"):
        chat = gr.TextArea(value=get_latest_chats,every=5,label="Chat History (Updated every 5 seconds)",lines=10)
        msg = gr.Textbox(label="Message to send",placeholder="Hello world!")
        clear = gr.ClearButton([msg, chat],value="Clear Chat History")

        def respond(message):
            if 'bot' in globals():
                bot.chat(message)
                return ""
            else:
                return ""
        
        def delete():
            chat_history.clear()
        
        clear.click(delete)
        msg.submit(respond, inputs=[msg],outputs=[msg])
        
    with gr.Tab("Plugins"):
        with gr.Accordion("Schematic Builder", open=False):
            
            def check():
                if not plugins.schematic in plugin_list:
                    return "The Schematic Builder plugin is not loaded"
                else:
                    return "pluginloaded"
               
            gr.Textbox(value=check, every=30, show_label=False)
            # with gr.Row():
            #     with gr.Column(scale=1, ):
            file_output = gr.File(file_types=[".schematic", ".nbt", ".schem"], label="Schematic File (.schematic .nbt .schem)",file_count="single")
            with gr.Row():
                with gr.Column(scale=1, ):
                    x = gr.Number(label="X Coordinate",info="The X coord to build at", precision=0)
                with gr.Column(scale=1, ):
                    z = gr.Number(label="Z Coordinate",info="The Z coord to build at", precision=0)
            # upload_button = gr.UploadButton("Click to Upload a schematic", file_count="single")
            # upload_button.upload(upload_file, upload_button, file_output)
            build = gr.Button("Build schematic", variant='primary')
            build.click(build_schematic, inputs=[file_output, x, z])
            
            
        with gr.Accordion("Build Cactus Farm", open=False):
            gr.Markdown("")
            
        with gr.Accordion("Auto Farm", open=False):
            with gr.Row():
                with gr.Column(scale=1, ):
                    crop_type = gr.Dropdown(["wheat_seeds", "wheat", "beetroot_seeds", "beetroot", "carrot", "potato", "poisonous_potato", "melon", "melon_slice", "melon_seeds", "melon_stem", "attached_melon_stem", "pumpkin", "carved_pumpkin", "pumpkin_seeds", "pumpkin_stem", "attached_pumpkin_stem", "torchflower_seeds", "torchflower_crop", "torchflower", "pitcher_pod", "pitcher_crop", "pitcher_plant", "farmland", "bamboo", "cocoa_beans", "sugar_cane", "sweet_berries", "cactus", "mushrooms", "kelp", "sea_pickle", "nether_wart", "chorus_fruit", "fungus", "glow_berries"], label="Crop Type",info="The Crop type to farm", interactive=True)
                with gr.Column(scale=1, ):
                    seed_name = gr.Dropdown(["wheat_seeds", "wheat", "beetroot_seeds", "beetroot", "carrot", "potato", "poisonous_potato", "melon", "melon_slice", "melon_seeds", "melon_stem", "attached_melon_stem", "pumpkin", "carved_pumpkin", "pumpkin_seeds", "pumpkin_stem", "attached_pumpkin_stem", "torchflower_seeds", "torchflower_crop", "torchflower", "pitcher_pod", "pitcher_crop", "pitcher_plant", "farmland", "bamboo", "cocoa_beans", "sugar_cane", "sweet_berries", "cactus", "mushrooms", "kelp", "sea_pickle", "nether_wart", "chorus_fruit", "fungus", "glow_berries"], label="Seed Name",info="The Seed name to plant back", interactive=True)
                with gr.Column(scale=1, ):
                    harvest_name = gr.Dropdown(["wheat_seeds", "wheat", "beetroot_seeds", "beetroot", "carrot", "potato", "poisonous_potato", "melon", "melon_slice", "melon_seeds", "melon_stem", "attached_melon_stem", "pumpkin", "carved_pumpkin", "pumpkin_seeds", "pumpkin_stem", "attached_pumpkin_stem", "torchflower_seeds", "torchflower_crop", "torchflower", "pitcher_pod", "pitcher_crop", "pitcher_plant", "farmland", "bamboo", "cocoa_beans", "sugar_cane", "sweet_berries", "cactus", "mushrooms", "kelp", "sea_pickle", "nether_wart", "chorus_fruit", "fungus", "glow_berries"], label="Harvest Name",info="The block name to harvest", interactive=True)
                
            with gr.Row():
                with gr.Column(scale=1, ):
                    x = gr.Number(label="X Coordinate",info="The X coord to build at", precision=0)
                with gr.Column(scale=1, ):
                    z = gr.Number(label="Z Coordinate",info="The Z coord to build at", precision=0)
            
            with gr.Row():
                with gr.Column(scale=1, ):
                    x = gr.Number(label="X Coordinate",info="The X coord to build at", precision=0)
                with gr.Column(scale=1, ):
                    z = gr.Number(label="Z Coordinate",info="The Z coord to build at", precision=0)
            with gr.Accordion("Optional Settings", open=False):
                gr.Markdown("Optional Settings")
            start_farm = gr.Button("Start auto Farming", variant='primary')
            
        with gr.Accordion("Discord Rich Presence", open=False):        
            with gr.Row():
                with gr.Column(scale=1, ):
                    state = gr.Textbox(label="State",info="The state to display")
                with gr.Column(scale=1, ):
                    details = gr.Textbox(label="Details",info="The details to display")
                with gr.Column(scale=1, ):
                    large_image = gr.Textbox(label="Large Image (url)",info="The large image to display")
                with gr.Column(scale=1, ):
                    large_text = gr.Textbox(label="Large Text",info="The large text to display")
                with gr.Column(scale=1, ):
                    small_image = gr.Textbox(label="Small Image (url)",info="The small image to display")
                with gr.Column(scale=1, ):
                    small_text = gr.Textbox(label="Small Text",info="The small text to display")
            
            def update_presence_def(state="No state provided", details="No details Provided", large_image=None, large_text=None, small_image=None, small_text=None):
                print(details)
                if 'bot' in globals():
                    bot.discordrp(state=state, details=details, start=time.time())
                else:
                    pass
            
            update_presence = gr.Button("Update Presence", variant='primary')
            update_presence.click(update_presence_def, inputs=[state, details, large_image, large_text, small_image, small_text])
            
    
    with gr.Tab("Movements"):
        with gr.Tab("Basic Movements"):
            with gr.Row():
                with gr.Column(scale=1, ):
                    jump = gr.Button("Start Jumping")
                    jump = gr.Button("Stop Jumping")
                with gr.Column(scale=1, ):
                    jump = gr.Button("Start walking forward")
                    jump = gr.Button("Stop walking forward")
        with gr.Tab("Follow Player/Entity"):
            gr.Markdown("")
                
            

    with gr.Tab("Player Info"):
        # refresh_button = gr.Button("Refresh")
        with gr.Row():
            with gr.Column(scale=1):
                health = gr.Textbox(value=get_player_health, label=f"Health", every=5)
            with gr.Column(scale=1):
                food = gr.Textbox(value=get_player_food, label=f"Food", every=5)
            with gr.Column(scale=1):
                experience = gr.Textbox(value=get_player_experience, label=f"Experience Level", every=5)
            with gr.Column(scale=1):
                difficulty = gr.Textbox(value=get_player_difficulty, label=f"Difficulty", every=5)
            with gr.Column(scale=1):
                all_data = gr.Textbox(value=get_all_data, label=f"All Data", every=5)
        # refresh_button.click(get_player_info, outputs=[health, food, experience])

    with gr.Tab("System Resources"):
        # refresh_button = gr.Button("Refresh")
        import psutil
        import platform
        
        def cpu():
            return psutil.cpu_percent(interval=5)
        
        def ram_used():
            return psutil.virtual_memory().percent
        
        def ram_available():
            
            ram=  psutil.virtual_memory().available / (1024.0 ** 3)
            return "{:.1f}".format(ram)
        
        
        with gr.Row():
            with gr.Column(scale=1):
                health = gr.Textbox(value=ram_used, label=f"Ram Used (%)", every=5)
            with gr.Column(scale=1):
                food = gr.Textbox(value=ram_available, label=f"Available Ram (GB)", every=5)
            with gr.Column(scale=1):
                experience = gr.Textbox(value=cpu, label=f"CPU usage (%)", every=5)
            with gr.Column(scale=1):
                difficulty = gr.Textbox(value=platform.system, label=f"System Type")
        # refresh_button.click(get_player_info, outputs=[health, food, experience])
if __name__ == "__main__":
    ui.queue().launch(server_port=8000, show_api=False, quiet=True)

    # gr.themes.builder()
    



import uvicorn
import lodestone
import structlog
from fastapi import FastAPI, Request
from fastapi.responses import JSONResponse
from rich.console import Console
import time 
from rich.panel import Panel
from rich import print
from rich.layout import Layout
app = FastAPI()
logger = structlog.get_logger()
console = Console()
layout = Layout()

def fastapi():

    print(Panel("Welcome to the Lodestone project!\n\nAPI docs are avable at: https://open-delivery-bot-documentation.vercel.app/", title="Welcome", expand=False, border_style="green"))
    
    logger.info("API running at http://localhost:5000/ (beta)")

    @app.middleware("http")
    async def add_process_time_header(request: Request, call_next):
        with console.status(f"[bold green][API] {request.method} {request.scope['path']} ...\n") as status:
            start_time = time.time()
            response = await call_next(request)
            process_time = time.time() - start_time
            # logger.warning(f"[API] {request.method} {request.scope['path']} ✅ {process_time} ms")
            status.update(f"[bold green][API] {request.method} {request.scope['path']} Done!\n")
            logger.info(f"[API] {request.method} {request.scope['path']}")
            return response

    @app.get("/")
    async def home():
        time.sleep(1)
        return JSONResponse(content="Welcome")
    


    @app.get("/api/v1/get_msa")
    async def startup(email:str = ""):
        global msa
        try:
            msa.stop()
        except:
            pass
        msa = lodestone.createBot(host="og-network.net", username=email, version="1.19", ls_api_mode=True)
        maxLoops = 0
        while msa.msa_status == False:
            if maxLoops >= 20:
                return JSONResponse(content={'user_code': "Already signed in"})
            time.sleep(1)
            maxLoops += 1
        if msa.msa_status == True:
            
            return JSONResponse(content=msa.msa_data)
        


    uvicorn.run(app, log_level="critical", port=5000)
    # uvicorn.run(app, port=5000)